package geom;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Observable;
import java.util.Observer;

import javax.vecmath.Vector2d;
import javax.vecmath.Vector3d;

public abstract class Curve extends PolyLine implements Observer {
	
	private double totalLength = 0;
	private ArrayList<Double> lengths = new ArrayList<Double>();
	public ArrayList<Vector3d> tangents = new ArrayList<Vector3d>();
	
	protected DPoint start;
	protected DPoint end;
	protected int subDiv;
	
	public Curve(Color color, DPoint start, DPoint end, int subDiv) {
		
		super(color, new ArrayList<Point>(), false);
		this.start = start;
		this.end = end;
		this.subDiv = subDiv;
		
		attachObservers();
		this.update(this, null);
	}
	
	public abstract ArrayList<Point> generatePoints();
	public abstract ArrayList<Vector3d> generateTangents();

	public void update(Observable o, Object arg) {
		
		ArrayList<Point> newPoints = generatePoints();
		
		this.points.clear();
		this.points.addAll(newPoints);
		
		ArrayList<Vector3d> newTangents = generateTangents();
		
		this.tangents.clear();
		this.tangents.addAll(newTangents);
		
		computeLength();
	}

	private void computeLength() {
		
		lengths.clear();
		totalLength = 0;
		
		for (int i = 0; i < this.points.size() - 1; i++) {
			
			Point p0 = this.points.get(i);
			Point p1 = this.points.get(i + 1);
			
			Vector3d dir = new Vector3d(p1.position);
			
			dir.sub(p0.position);
			double length = dir.length();
			this.lengths.add(length);
			
			this.totalLength += length;
		}
	}
	
	private double computeLengthToPoint(Point p) {
		
		if ( !this.points.contains(p)) return Double.MAX_VALUE;
		
		double length = 0;
		
		for (int i = 0; i < this.points.size(); i++) {
			
			if ( p == this.points.get(i)) return length;
			
			length += this.lengths.get(i);
		}
		
		return Double.MAX_VALUE;
	}
	
	public Vector3d nearestPoint(Vector3d given, Vector2d distances) {
		
		Vector3d dir = new Vector3d();
		
		double minDistance = Double.MAX_VALUE;	
		Point p0 = null;
		int index = 0;
		
		int i = 0;
		
		for (Point point : this.points) {
			
			dir.set(given);
			dir.sub(point.position);
			double distance = dir.length();
			
			if ( distance < minDistance) {
				
				minDistance = distance;
				p0 = point;
				index = i;
			}
			i++;
		}
		
		distances.x = computeLengthToPoint(p0);
		
		Point p1 = null;
		double sign = 1;
		double distance = Double.MAX_VALUE;
		
		if ( index > 0 ) {
			
			p1 = this.points.get(index - 1);
			dir.set(p1.position);
			dir.sub(given);
			distance = dir.length();
			sign = -1;
		}
		
		if ( index < this.points.size() - 1) {
			
			Point next = this.points.get(index + 1);
			dir.set(next.position);
			dir.sub(given);
			if ( distance > dir.length() ) {
				
				p1 = next;
				sign = 1;
			}
		}
		
		dir.set(p1.position);
		dir.sub(p0.position);
		
		Vector3d dir2 = new Vector3d(p0.position);
		dir2.sub(given);
		
		double t = - dir2.dot(dir) / dir.lengthSquared();
		
		if ( t > 1 ) return p1.position;
		else if ( t < 0 ) return p0.position;
		
		dir.scale(t);
		distances.x += dir.length() * sign;
		dir.add(p0.position);
		
		return dir;
	}
	
	public Vector3d pointAtDistance(double distance, Vector3d tangent) {
		
		double tmpDistance = 0;
		
		for (int i = 0; i < this.lengths.size(); i++) {
			
			tmpDistance += lengths.get(i);
			
			if ( tmpDistance > distance) {
				
				Point p0 = this.points.get(i);
				Point p1 = this.points.get(i + 1);
				
				double t = (distance - tmpDistance + lengths.get(i)) / lengths.get(i);
				
				Vector3d result = new Vector3d();			
				result.interpolate(p0.position, p1.position, t);
				
				Vector3d t0 = this.tangents.get(i);
				Vector3d t1 = this.tangents.get(i + 1);
				
				tangent.interpolate(t0, t1, t);
				
				return result;
			}
		}
		
		return new Vector3d(this.points.get(this.points.size() - 1).position);
	}

	public double getTotalLength() {
		return totalLength;
	}

	@Override
	public void attachObservers() {
		
		start.addObserver(this);
		end.addObserver(this);
	}
}
