package shapes;

import java.awt.Color;
import java.awt.Graphics;
import java.util.ArrayList;
import java.util.Observable;
import java.util.Observer;

import javax.vecmath.Matrix4d;
import javax.vecmath.Vector2d;
import javax.vecmath.Vector3d;

import core.Core;
import core.Utils;

import drawing.Catchable;
import drawing.Enviroment;

import geom.Arc;
import geom.Line;
import geom.Point;
import geom.Rectangle;
import geom.RotateGroup;
import geom.Shape;
import geom.TranslateGroup;

public class Pattern extends Shape implements Catchable, Observer {

	public Road road;
	
	private RotateGroup rg;
	private TranslateGroup tg;
	private Line distanceLine;
	
	public double distance;
	public ArrayList<Arc> arcs = new ArrayList<Arc>();
	public double throttle = 0.5;
	public double steer = 0.5;

	public Pattern(Road road, double startX, double startY, double velocityX, double velocityY) {
		
		Rectangle r = new Rectangle(Color.BLACK, Color.GRAY, 0, 0, 5, 2);
    	r.setZIndex(1);
		
		this.rg = new RotateGroup(velocityX, velocityY);
		this.tg = new TranslateGroup(startX, startY);
		
		this.road = road;
		this.rg.addShape(r);
		this.tg.addShape(rg);
		
		this.rg.addObserver(this);
		this.tg.addObserver(this);
		
		this.distanceLine = new Line(new Point(0,0), new Point(0, 0));
		
		recomputeHelpers();
	}
	
	@Override
	public void draw(Graphics g) {
		
		if ( !Core.instance.isRoadEditing ) {
			
			for (Arc arc : this.arcs) {

				arc.draw(g);
			}
			
			this.tg.draw(g);
			this.distanceLine.draw(g);
		}
	}

	public void transform(Matrix4d windowTransform) {
		
		if ( !Core.instance.isRoadEditing ) {
			
			this.tg.transform(windowTransform);		
			this.distanceLine.transform(windowTransform);
			
			for (Arc arc : this.arcs) {
				
				arc.transform(windowTransform);
			}
		}
	}
	
	private void recomputeHelpers() {
		
		Vector2d distances = new Vector2d();
		
		this.distanceLine.p1.position.set(this.tg.referencePoint.position);
		this.distanceLine.p2.position.set(this.road.nearestPoint(this.distanceLine.p1.position, distances));
		
		arcs.clear();
		double step = Enviroment.VISIBILITY / Enviroment.NUM_OF_ANGLES;
		double basicDistance = distances.x;
		Vector3d xAxis = new Vector3d(1,0,0);
		
		//System.out.println("----------------------");
		
		for ( int i = 0; i < Enviroment.NUM_OF_ANGLES; i++) {
			
			Vector3d tangent = new Vector3d();
			Vector3d velocity = new Vector3d(this.getVelocity());
			velocity.normalize();
			Vector3d pointAtDistance = this.road.pointAtDistance(basicDistance + step * i, tangent);
			Arc arc = new Arc(Color.DARK_GRAY, "fi"+i, pointAtDistance, tangent, velocity);
			
			arc.startAngle = Utils.angle(xAxis, tangent);
			arc.arcAngle = Utils.angle(tangent, velocity);
			
			if ( i == 0) {
				
				Vector3d direction = new Vector3d(this.distanceLine.p1.position);
				direction.sub(this.distanceLine.p2.position);
				
				if ( Utils.angle(tangent, direction) > 0) {
					
					this.distance = direction.length();
				}
				else {
					
					this.distance = -direction.length();
				}
			}
			
			this.arcs.add(arc);
		}
		
		Core.instance.mainFrame.update(null, null);
	}
	
	public void update(Observable o, Object arg) {
		
		recomputeHelpers();
	}

	public Vector3d getVelocity() {
		
		return this.rg.referencePoint.position;
	}
	 
	public void setVelocity(Vector3d velocity) {
		
		this.rg.referencePoint.setPosition(velocity);
	}
	
	public Vector3d getPosition() {
		
		return this.tg.referencePoint.position;
	}
	
	public void setPosition(Vector3d position) {
		
		this.tg.referencePoint.setPosition(position);
	}
	
	public Catchable isCatch(double x, double y) {

		if ( !Core.instance.isRoadEditing ) {
			
			return this.tg.isCatch(x, y);
		}
		else return null;
	}

	@Override
	public String toString() {

		return "(" + road.name + ", " + Enviroment.NUM_FORMAT.format(Utils.msToKmh(this.getVelocity().length())) + ", " +
			Enviroment.NUM_FORMAT.format(this.distance) + ") -> (" +
			Enviroment.NUM_FORMAT.format(this.throttle) + ", " + Enviroment.NUM_FORMAT.format(this.steer) + ")";
	}

	@Override
	public void attachObservers() {
		
		this.rg.attachObservers();
		this.rg.update(null, null);
		this.tg.attachObservers();
		this.tg.update(null, null);
		
		this.rg.addObserver(this);
		this.tg.addObserver(this);
	}
}
