package shapes;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.util.ArrayList;

import javax.vecmath.Matrix4d;
import javax.vecmath.Vector2d;
import javax.vecmath.Vector3d;

import core.Core;

import drawing.Catchable;
import drawing.Enviroment;

import geom.Bezier;
import geom.Curve;
import geom.DPoint;
import geom.Point;
import geom.Shape;

public class Road extends Shape implements Catchable {
	
	public String name; 
	public ArrayList<DPoint> points = new ArrayList<DPoint>();
	private ArrayList<Curve> curves = new ArrayList<Curve>();
	
	public Road(Color color, String name) {
	
		super(color);
		this.name = name;
	}

	@Override
	public void draw(Graphics g) {
		
		Graphics2D g2 = (Graphics2D)g;
		
		Stroke stroke = g2.getStroke();

		g2.setStroke(new BasicStroke(Enviroment.ROAD_WIDTH, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
		
		g.setColor(this.getColor());
		
		for (Curve c : this.curves) {
			
			c.draw(g);
		}
		
		g2.setStroke(stroke);
		
		for (Curve c : this.curves) {
			
			g.setColor(c.getColor());
			c.draw(g);
		}
		
		if ( Core.instance.isRoadEditing ) {
		
			for (DPoint point : points) {

				point.draw(g);
			}
		}
	}

	public void transform(Matrix4d windowTransform) {
		
		for (Curve c : this.curves) {
			
			c.transform(windowTransform);
		}
		
		for (DPoint point : points) {
			
			point.transform(windowTransform);
		}
	}

	public Catchable isCatch(double x, double y) {
		
		if ( Core.instance.isRoadEditing ) {
		
			for (DPoint point : points) {

				Catchable c = point.isCatch(x, y);

				if ( c != null ) return c;
			}
		}
		return null;
	}
	
	public void addPoint(DPoint point) {
		
		if ( this.points.size() > 0 ) {
			
			Curve c = 
				new Bezier(Color.BLUE, this.points.get(this.points.size() - 1), point, Enviroment.CURVE_SEGMENTS);			
			this.curves.add(c);
		}
		this.points.add(point);
	}
	
	public void removeLastPoint() {
		
		if ( this.curves.size() > 0 ) {
			
			this.curves.remove(this.curves.size() - 1);
		}
		if ( this.points.size() > 0 ) {
			
			this.points.remove(this.points.size() - 1);
		}
	}

	@Override
	public String toString() {
		// TODO Auto-generated method stub
		return this.name;
	}

	public ArrayList<DPoint> getPoints() {
		return points;
	}

	public ArrayList<Curve> getCurves() {
		return curves;
	}

	@Override
	public void attachObservers() {
		
		for (DPoint point : this.points) {
		
			point.attachObservers();
			point.update(null, null);
		}
		
		for (Curve curve : this.curves) {
			
			curve.attachObservers();
			curve.update(null, null);
		}
	}
	
	public Vector3d nearestPoint(Vector3d given, Vector2d distances) {
		
		Vector3d dir = new Vector3d();
		
		double minDistance = Double.MAX_VALUE;	
		Vector3d result = null;
		double distanceFromBeginToResult = 0;
		double distanceFromBegin = 0;
		
		for (Curve curve : this.curves) {
			
			Vector2d segmentDistances = new Vector2d();
			Vector3d nearest = curve.nearestPoint(given, segmentDistances);
			
			dir.set(given);
			dir.sub(nearest);
			double distance = dir.length();
			
			if ( distance < minDistance) {
				
				minDistance = distance;
				result = nearest;
				distanceFromBeginToResult = distanceFromBegin + segmentDistances.x;
			}
			
			distanceFromBegin += curve.getTotalLength();
		}
		
		distances.x = distanceFromBeginToResult;
		return result;
	}
	
	public Vector3d pointAtDistance(double distance, Vector3d tangent) {
		
		double tmpDistance = 0;
		
		for (Curve curve : this.curves) {
			
			tmpDistance += curve.getTotalLength();
			
			if ( tmpDistance > distance ) {
				
				return curve.pointAtDistance(distance - tmpDistance + curve.getTotalLength(), tangent);
			}
		}
		
		Curve lastCurve = this.curves.get(this.curves.size() - 1);
		tangent.set(lastCurve.tangents.get(lastCurve.tangents.size() - 1));
		return this.points.get(this.points.size() - 1).position;
	}
}
